from lib2to3.fixer_base import BaseFix
from lib2to3.fixer_util import LParen, RParen, find_indentation
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols
from lib2to3.pytree import Leaf, Node
from textwrap import TextWrapper

from .utils import tuplize_comments, get_quotes, wrap_leaves, first_child_leaf

MAX_CHARS = 79
OPENING_TOKENS = [token.LPAR, token.LSQB, token.LBRACE]
CLOSING_TOKENS = [token.RPAR, token.RSQB, token.RBRACE]
SYMBOLS_WITH_NEWLINES_IN_COLONS = [symbols.funcdef, symbols.classdef,
    symbols.if_stmt, symbols.for_stmt, symbols.while_stmt, symbols.lambdef,
    symbols.try_stmt, symbols.with_stmt]


class FixMaximumLineLength(BaseFix):
    u'''
    Limit all lines to a maximum of 79 characters.

    There are still many devices around that are limited to 80 character
    lines; plus, limiting windows to 80 characters makes it possible to have
    several windows side-by-side.  The default wrapping on such devices looks
    ugly.  Therefore, please limit all lines to a maximum of 79 characters.
    For flowing long blocks of text (docstrings or comments), limiting the
    length to 72 characters is recommended.
    '''

    explicit = True  # The user must ask for this fixer

    def match(self, node):
        if (node.type in [token.NEWLINE] or node.type == token.COLON and node.
            parent.type in SYMBOLS_WITH_NEWLINES_IN_COLONS):
            # Sometimes the newline is wrapped into the next node, so we need
            # to check the colons also.
            if self.need_to_check_node(node):
                # For colon nodes, we need to add the len of the colon also
                return True
        if any(len(line) > MAX_CHARS for line in node.prefix.split('\n')):
            # There is a line in the prefix greater than MAX_CHARS
            return True
        return False

    def transform(self, node, results):
        if self.node_needs_splitting(node):
            node_to_split = node.prev_sibling
            if node_to_split.type == token.STRING:
                self.fix_docstring(node_to_split)
            else:
                if isinstance(node_to_split, Leaf):
                    node_to_split = node_to_split.parent
                combined_prefix = self.fix_leaves(node_to_split)
                if combined_prefix:
                    node.prefix = u"%s\n%s" % (node.prefix, combined_prefix.
                        rstrip())
        if (any(len(line) > MAX_CHARS for line in node.prefix.split(u'\n')) or
            node.prefix.count(u"#") and node.column + len(node.prefix) >
            MAX_CHARS):
            # Need to fix the prefix
            self.fix_prefix(node)

    @staticmethod
    def need_to_check_node(node):
        # Returns if the node or it's docstring might need to be split
        if node.column > MAX_CHARS:
            return True
        if (node.type == token.COLON
            and node.column + len(node.value) > MAX_CHARS):
            return True
        if node.prev_sibling and any(child.column + len(child.value)
            > MAX_CHARS for child in node.prev_sibling.leaves()):
            return True

    @staticmethod
    def node_needs_splitting(node):
        if not node.prev_sibling:
            return False

        if node.type == token.NEWLINE:
            node_length = len(node.prefix)
        elif node.type == token.COLON:
            node_length = len(node.prefix) - len(node.value)
        if node.type in [token.NEWLINE, token.COLON]:
            if node.column - node_length > MAX_CHARS:
                return True

            for child in node.prev_sibling.leaves():
                if child.type == token.STRING:
                    lines = node.value.split(u'\n')
                    if child.column + len(lines.pop(0)) > MAX_CHARS:
                        return True
                    elif any(len(line) > MAX_CHARS for line in lines):
                        return True
                elif child.column + len(child.value) > MAX_CHARS:
                    return True

    def fix_prefix(self, node):
        before_comments, comments, after_comments = tuplize_comments(node.
            prefix)

        # Combine all comment lines together
        all_comments = u' '.join([line.replace(u'#', u'', 1).lstrip() for line
            in comments.split(u'\n')])

        # It's an inline comment if it has not newlines
        is_inline_comment = not node.prefix.count(u'\n')

        initial_indent_level = comments.find(u'#')
        if initial_indent_level == -1:
            split_lines = [u'']
        else:
            if is_inline_comment and node.prev_sibling:
                # If inline comment, find where the prev sibling started to
                # know how to indent lines
                initial_indent_level = (first_child_leaf(node.prev_sibling).
                    column)
            indent = u'%s# ' % (u' ' * initial_indent_level)

            wrapper = TextWrapper(width=MAX_CHARS, initial_indent=indent,
                subsequent_indent=indent)
            split_lines = wrapper.wrap(all_comments)

            if is_inline_comment:
                # If inline comment is too long, we'll move it to the next line
                split_lines[0] = u"\n%s" % split_lines[0]
            else:
                # We need to add back a newline that was lost above
                after_comments = u"\n%s" % after_comments
        new_prefix = u'%s%s%s' % (before_comments, u'\n'.join(split_lines),
            after_comments.lstrip(u' '))
        # Append the trailing spaces back
        if node.prefix != new_prefix:
            node.prefix = new_prefix
            node.changed()

    def fix_docstring(self, node_to_split):
        # docstrings
        quote_start, quote_end = get_quotes(node_to_split.value)
        max_length = MAX_CHARS - node_to_split.column

        triple_quoted = quote_start.count(u'"""') or quote_start.count(u"'''")
        comment_indent = u' ' * (4 + node_to_split.column)

        if not triple_quoted:
            # If it's not tripled-quoted, we need to start and end each line
            # with quotes
            comment_indent = u'%s%s' % (comment_indent, quote_start)
            # Since we will be appending the end_quote after each line after
            # the splitting
            max_length -= len(quote_end)
            # If it's not triple quoted, we need to paren it
            node_to_split.value = u"(%s)" % node_to_split.value

        wrapper = TextWrapper(width=max_length,
            subsequent_indent=comment_indent)
        split_lines = wrapper.wrap(node_to_split.value)

        if not triple_quoted:
            # If it's not triple quoted, we need to close each line except for
            # the last one
            split_lines = [u"%s%s" % (line, quote_end) if index != len(
                split_lines) - 1 else line for index, line in enumerate(
                split_lines)]

        new_nodes = [Leaf(token.STRING, split_lines.pop(0))]
        for line in split_lines:
            new_nodes.extend([Leaf(token.NEWLINE, u'\n'), Leaf(token.STRING,
                line)])

        node_to_split.replace(new_nodes)
        node_to_split.changed()

    def fix_leaves(self, node_to_split):
        parent_depth = find_indentation(node_to_split)
        new_indent = u"%s%s" % (u' ' * 4, parent_depth)
        # For now, just indent additional lines by 4 more spaces

        child_leaves = []
        combined_prefix = u""
        prev_leaf = None
        for index, leaf in enumerate(node_to_split.leaves()):
            if index and leaf.prefix.count(u'#'):
                if not combined_prefix:
                    combined_prefix = u"%s#" % new_indent
                combined_prefix += leaf.prefix.split(u'#')[-1]

            # We want to strip all newlines so we can properly insert newlines
            # where they should be
            if leaf.type != token.NEWLINE:
                if leaf.prefix.count(u'\n') and index:
                    # If the line contains a newline, we need to strip all
                    # whitespace since there were leading indent spaces
                    if (prev_leaf and prev_leaf.type in [token.DOT, token.LPAR]
                        or leaf.type in [token.RPAR]):
                        leaf.prefix = u""
                    else:
                        leaf.prefix = u" "

                    # Append any trailing inline comments to the combined
                    # prefix
                child_leaves.append(leaf)
                prev_leaf = leaf

        # Like TextWrapper, but for nodes. We split on MAX_CHARS - 1 since we
        # may need to insert a leading parenth. It's not great, but it would be
        # hard to do properly.
        split_leaves = wrap_leaves(child_leaves, width=MAX_CHARS - 1,
            subsequent_indent=new_indent)
        new_node = Node(node_to_split.type, [])

        # We want to keep track of if we are breaking inside a parenth
        open_count = 0
        need_parens = False
        for line_index, curr_line_nodes in enumerate(split_leaves):
            for node_index, curr_line_node in enumerate(curr_line_nodes):
                if line_index and not node_index:
                    # If first node in non-first line, reset prefix since there
                    # may have been spaces previously
                    curr_line_node.prefix = new_indent
                new_node.append_child(curr_line_node)
                if curr_line_node.type in OPENING_TOKENS:
                    open_count += 1
                if curr_line_node.type in CLOSING_TOKENS:
                    open_count -= 1

            if line_index != len(split_leaves) - 1:
                # Don't add newline at the end since it it part of the next
                # sibling
                new_node.append_child(Leaf(token.NEWLINE, u'\n'))

                # Checks if we ended a line without being surrounded by parens
                if open_count <= 0:
                    need_parens = True
        if need_parens:
            # Parenthesize the parent if we're not inside parenths, braces,
            # brackets, since we inserted newlines between leaves
            parenth_before_equals = Leaf(token.EQUAL, u"=") in split_leaves[0]
            self.parenthesize_parent(new_node, parenth_before_equals)
        node_to_split.replace(new_node)

        return combined_prefix

    def parenthesize_parent(self, node_to_split, parenth_before_equals):
        if node_to_split.type == symbols.print_stmt:
            self.parenthesize_print_stmt(node_to_split)
        elif node_to_split.type == symbols.return_stmt:
            self.parenthesize_after_arg(node_to_split, u"return")
        elif node_to_split.type == symbols.expr_stmt:
            if parenth_before_equals:
                self.parenthesize_after_arg(node_to_split, u"=")
            else:
                self.parenthesize_expr_stmt(node_to_split)
        elif node_to_split.type == symbols.import_from:
            self.parenthesize_after_arg(node_to_split, u"import")
        elif node_to_split.type in [symbols.power, symbols.atom]:
            self.parenthesize_call_stmt(node_to_split)
        elif node_to_split.type in [symbols.or_test, symbols.and_test, symbols
            .not_test, symbols.test, symbols.arith_expr, symbols.comparison]:
            self.parenthesize_test(node_to_split)
        elif node_to_split.type == symbols.parameters:
            # Paramteres are always parenthesized already
            pass

    def parenthesize_test(self, node_to_split):
        first_child = node_to_split.children[0]
        if first_child != LParen():
            # node_to_split.children[0] is the "print" literal strip the
            # current 1st child, since we will be prepending an LParen
            if first_child.prefix != first_child.prefix.strip():
                first_child.prefix = first_child.prefix.strip()
                first_child.changed()
            left_paren = LParen()
            left_paren.prefix = u" "
            node_to_split.insert_child(0, left_paren)
            node_to_split.append_child(RParen())
            node_to_split.changed()

    def parenthesize_print_stmt(self, node_to_split):
        # print "hello there"
        # return a, b
        second_child = node_to_split.children[1]
        if second_child != LParen():
            # node_to_split.children[0] is the "print" literal strip the
            # current 1st child, since we will be prepending an LParen
            if second_child.prefix != second_child.prefix.strip():
                second_child.prefix = second_child.prefix.strip()
                second_child.changed()
            node_to_split.insert_child(1, LParen())
            node_to_split.append_child(RParen())
            node_to_split.changed()

    def parenthesize_after_arg(self, node_to_split, value):
        # parenthesize the leaves after the first node with the value
        value_index = 0
        for index, child in enumerate(node_to_split.children):
            if child.value == value:
                value_index = index + 1
                break
        value_child = node_to_split.children[value_index]
        if value_child != LParen():
            # strip the current 1st child, since we will be prepending an
            # LParen
            if value_child.prefix != value_child.prefix.strip():
                value_child.prefix = value_child.prefix.strip()
                value_child.changed()
            # We set a space prefix since this is after the '='
            left_paren = LParen()
            left_paren.prefix = u" "
            node_to_split.insert_child(value_index, left_paren)
            node_to_split.append_child(RParen())
            node_to_split.changed()

    def parenthesize_expr_stmt(self, node_to_split):
        # x = "foo" + bar
        if node_to_split.children[0] != LParen():
            node_to_split.insert_child(0, LParen())
            node_to_split.append_child(RParen())
            node_to_split.changed()

    def parenthesize_call_stmt(self, node_to_split):
        # a.b().c()
        first_child = node_to_split.children[0]
        if first_child != LParen():
            # Since this can be at the beginning of a line, we can't just
            # strip the prefix, we need to keep leading whitespace
            first_child.prefix = u"%s(" % first_child.prefix
            first_child.changed()
            node_to_split.append_child(RParen())
            node_to_split.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols
from lib2to3.pytree import Leaf

from .utils import OPERATORS, UNARY_OPERATORS

ARG_SYMBOLS = [symbols.arglist, symbols.varargslist, symbols.typedargslist]
KEYWORKD_ARG_SYMBOLS = [symbols.argument, symbols.arglist, symbols.
    typedargslist]


class FixWhitespaceAroundOperator(BaseFix):
    u'''
    Avoid extraneous whitespace in the following situations:

    - More than one space around an assignment (or other) operator to
      align it with another.
    '''

    def match(self, node):
        if isinstance(node, Leaf) and node.value in OPERATORS:
            return True
        return False

    def transform(self, node, results):
        # Allow unary operators: -123, -x, +1.
        if (node.value in UNARY_OPERATORS and node.parent.type == symbols.
            factor):
            self.rstrip(node)
        # Allow argument unpacking: foo(*args, **kwargs).
        elif(node.value in [u'*', u'**'] and node.parent.type in ARG_SYMBOLS
            and (not node.prev_sibling or node.prev_sibling.type == token.
            COMMA)):
            self.rstrip(node)
        # Allow keyword assignment: foobar(foo=bar)
        elif node.value == u'=' and node.parent.type in KEYWORKD_ARG_SYMBOLS:
            self.no_spaces(node)
        # Finally check if the spacing actually needs fixing
        elif(node.prefix != u" " or node.get_suffix() != u" "):
            self.spaces(node)

    def rstrip(self, node):
        next_sibling = node.next_sibling
        next_sibling_new_prefix = next_sibling.prefix.lstrip(u' \t')
        if next_sibling.prefix != next_sibling_new_prefix:
            next_sibling.prefix = next_sibling_new_prefix
            next_sibling.changed()

    def no_spaces(self, node):
        if node.prefix != u"":
            node.prefix = u""
            node.changed()

        next_sibling = node.next_sibling
        next_sibling_new_prefix = next_sibling.prefix.lstrip(u' \t')
        if next_sibling.prefix != next_sibling_new_prefix:
            next_sibling.prefix = next_sibling_new_prefix
            next_sibling.changed()

    def spaces(self, node):
        if not node.prefix.count(u'\n'):
            # If there are newlines in the prefix, this is a continued line,
            # don't strip anything
            new_prefix = u" %s" % node.prefix.lstrip(u' \t')
            if node.prefix != new_prefix:
                node.prefix = new_prefix
                node.changed()

        next_sibling = node.next_sibling
        if not next_sibling:
            return
        if next_sibling.prefix.count(u'\n'):
            next_sibling_new_prefix = next_sibling.prefix.lstrip(u' \t')
        else:
            next_sibling_new_prefix = u" %s" % next_sibling.prefix.lstrip(
                u' \t')
        if next_sibling.prefix != next_sibling_new_prefix:
            next_sibling.prefix = next_sibling_new_prefix
            next_sibling.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.fixer_util import Newline
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols


class FixMissingWhitespace(BaseFix):
    u'''
    Each comma, semicolon or colon should be followed by whitespace.
    '''

    def match(self, node):
        if (node.type in (token.COLON, token.COMMA, token.SEMI) and node.
            get_suffix() != u" "):
            # If there is a newline after, no space
            if (node.get_suffix().find(u'\n') == 0 or
                (node.next_sibling and node.next_sibling.children and
                node.next_sibling.children[0] == Newline())):
                return False
            # If we are using slice notation, no space necessary
            if node.parent.type in [symbols.subscript, symbols.sliceop]:
                return False
            # If a tuple with a single element, no space
            if not node.next_sibling:
                return False
            return True
        return False

    def transform(self, node, results):
        new_prefix = u" %s" % node.next_sibling.prefix.lstrip(u' \t')
        if node.next_sibling.prefix != new_prefix:
            node.next_sibling.prefix = new_prefix
            node.next_sibling.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pygram import python_symbols as symbols

from .utils import get_leaves_after_last_newline


class FixTrailingBlankLines(BaseFix):
    u'''
    Trailing blank lines are superfluous.
    '''

    def match(self, node):
        # We only want to work with the top-level input since this should only
        # run once.
        if node.type != symbols.file_input:
            return

        leaves_after_last_newline = get_leaves_after_last_newline(node)
        # Return any leaves with newlines.
        return [leaf for leaf in leaves_after_last_newline
                    if leaf.prefix.count('\n')]

    def transform(self, node, results):
        for index, result in enumerate(results):
            if index:
                # We've already stripped one newline. Strip any remaining
                if result.prefix != result.prefix.rstrip():
                    result.prefix = result.prefix.rstrip()
                    result.changed()
            else:
                # We haven't stripped any newlines yet. We need to strip all
                # whitespace, but leave a single newline.
                if result.prefix.strip():
                    # If there are existing comments, we need to add two
                    # newlines in order to have a trailing newline.
                    new_prefix = u'%s\n\n' % result.prefix.rstrip()
                else:
                    new_prefix = u'%s\n' % result.prefix.rstrip()
                if result.prefix != new_prefix:
                    result.prefix = new_prefix
                    result.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pygram import python_symbols as symbols

from .utils import get_leaves_after_last_newline


class FixMissingNewline(BaseFix):
    u'''
    The last line should have a newline.

    This is somewhat tricky since the parse tree
    sometimes categorizes newlines as token.DEDENTs
    '''

    def match(self, node):
        # We only want to work with the top-level input since this should only
        # run once.
        if node.type != symbols.file_input:
            return

        leaves_after_last_newline = get_leaves_after_last_newline(node)
        if not any(leaf.prefix.count('\n')
                   for leaf in leaves_after_last_newline):
            # If none of those have a prefix containing a newline,
            # we need to add one
            return leaves_after_last_newline[0]

    def transform(self, node, leaf):
        if leaf.prefix != u'\n':
            leaf.prefix = u'\n'
            leaf.changed()



from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols
from lib2to3.pytree import Leaf

BINARY_OPERATORS = frozenset(['**=', '*=', '+=', '-=', '!=', '<>',
    '%=', '^=', '&=', '|=', '==', '/=', '//=', '<=', '>=', '<<=', '>>=',
    '%', '^', '&', '|', '=', '/', '//', '<', '>', '<<'])
UNARY_OPERATORS = frozenset(['>>', '**', '*', '+', '-'])
OPERATORS = BINARY_OPERATORS | UNARY_OPERATORS
MAX_CHARS = 79


def get_leaves_after_last_newline(node):
    # Get all of the leaves after the last newline leaf
    all_leaves = []
    last_newline_leaf_index = -1
    for index, leaf in enumerate(node.leaves()):
        all_leaves.append(leaf)
        if leaf.type == token.NEWLINE:
            last_newline_leaf_index = index
    return all_leaves[last_newline_leaf_index + 1:]


def first_child_leaf(node):
    if isinstance(node, Leaf):
        return node
    elif node.children:
        return first_child_leaf(node.children[0])
    else:
        return None


def get_whitespace_before_definition(node):
    if node.prev_sibling:
        return get_last_child_with_whitespace(node.prev_sibling)


def get_last_child_with_whitespace(node):
    leaves = []
    for leaf in node.leaves():
        leaves.append(leaf)
    reverse_leaves = reversed(leaves)
    for leaf in reverse_leaves:
        if u'\n' in leaf.prefix or leaf.value == u'\n':
            return leaf


def has_parent(node, symbol_type):
    # Returns if node has a parent of type symbol_type
    if node.parent:
        return node.parent.type == symbol_type or has_parent(node.parent,
            symbol_type)


def prefix_indent_count(node):
    # Find the number of spaces preceding this line
    return len(node.prefix.split(u'\n')[-1].replace(u'\t', u' ' * 4))


def node_length(*nodes):
    return sum(len(node.prefix.strip(u'\n\t')) +
        len(node.value.strip(u'\n\t')) for node in nodes)


def tuplize_comments(prefix):
    # This tuplizes the newlines before and after the prefix
    # Given u'\n\n\n    # test comment\n    \n'
    # returns ([u'\n\n\n'], [u'    # test comment\n'], [u'    \n'])

    if not prefix:
        return (u'', u'', u'')

    # If there are no newlines, this was just a trailing comment. Leave it
    # alone.
    if not prefix.count(u'\n'):
        return (u'', prefix, u'')

    if prefix.count("#"):
        whitespace_before_first_comment = prefix[:prefix.index(u"#")]
        start_of_comment = whitespace_before_first_comment.rfind(u'\n')
        if prefix.count(u'\n') and not prefix.split(u'\n')[-1].strip():
            # Add a single newline back if there was a newline in the ending
            # whitespace
            comments = u"%s\n" % prefix[start_of_comment + 1:].rstrip()
        else:
            comments = prefix[start_of_comment + 1:].rstrip()
    else:
        if prefix.count(u'\n'):
            comments = prefix.rsplit(u'\n')[1]
            # If no comments, there are no comments except the trailing spaces
            # before the current line
        else:
            comments = prefix
    comments_start = prefix.index(comments)
    return prefix[:comments_start].strip(u' '), comments, prefix[
        comments_start + len(comments):]


def get_quotes(text):
    # Returns the quote type start and end
    # Given u"ur'the string'" returns (u"ur'", u"'")

    if text[:2].lower() in [u'br', u'ur']:
        leading_chars = 2
    elif text[:1].lower() in [u'b', u'u', u'r']:
        leading_chars = 1
    else:
        leading_chars = 0

    if text[leading_chars:leading_chars + 3] in [u'"""', u"'''"]:
        # Triple-quoted string
        quote_start = text[:leading_chars + 3]
    else:
        # Single-quoted string
        quote_start = text[:leading_chars + 1]
    return (quote_start, quote_start[leading_chars:])


# Like TextWrapper, but for leaves
def wrap_leaves(nodes, width=MAX_CHARS, initial_indent=u'',
    subsequent_indent=u''):
    lines = []

    # Fake the prefix of the first node to be the indent that it should be.
    # We'll set it back afterward.
    first_node_prefix = nodes[0].prefix
    nodes[0].prefix = u' ' * nodes[0].column

    nodes.reverse()
    while nodes:
        tracking_back = False
        curr_line = []
        curr_len = 0

        # Figure out which static string will prefix this line.
        if lines:
            indent = subsequent_indent
        else:
            indent = initial_indent

        # Maximum width for this line.
        curr_width = width - len(indent)

        while nodes:
            last_node = nodes[-1]

            if lines and not curr_line:
                # Strip prefixes for subsequent lines
                last_node.prefix = u''

            curr_node_length = node_length(last_node)

            # Can at least squeeze this chunk onto the current line.
            if curr_len + curr_node_length <= curr_width:
                curr_line.append(nodes.pop())
                curr_len += curr_node_length

            # Nope, this line is full.
            else:
                # only disallow breaking on/after equals if parent of this type
                if nodes and nodes[-1].type in [token.COMMA, token.EQUAL]:
                    # We don't want the next line to start on one of these
                    # tokens
                    tracking_back = True
                    nodes.append(curr_line.pop())
                if (curr_line and curr_line[-1].type == token.EQUAL and
                    curr_line[-1].parent.type != symbols.expr_stmt):
                    # We don't want this line to end on one of these tokens.
                    # Move the last two nodes back onto the list
                    tracking_back = True
                    nodes.extend(reversed(curr_line[-2:]))
                    del curr_line[-2:]
                break

        # The current line is full, and the next chunk is too big to fit on
        # *any* line (not just this one).
        if nodes:
            next_chunk_length = node_length(nodes[-1])
            if tracking_back:
                next_chunk_length += node_length(nodes[-2])
            if next_chunk_length > curr_width:
                curr_line.append(nodes.pop())
                if nodes and nodes[-1].type in [token.COMMA, token.EQUAL]:
                    # We don't want the next line to start on these chars, just
                    # add them here Check maximum_line_length3_in:4 for an
                    # example
                    curr_line.append(nodes.pop())
            elif (len(nodes) > 2 and not curr_line and
                node_length(*nodes[-3:]) > curr_width):
                # This scenario happens when we were not able to break on an
                # assignment statement above and the next line is still too
                # long. Remove the last 3 nodes and move them to curr_line
                curr_line.extend(reversed(nodes[-3:]))
                del nodes[-3:]
                if nodes and nodes[-1].type in [token.COMMA, token.EQUAL]:
                    curr_len += node_length(nodes[-1])
                    curr_line.append(nodes.pop())

        if curr_line:
            curr_line[0].prefix = "%s%s" % (indent, curr_line[0].prefix)
            lines.append(curr_line)
        else:
            assert False, ("There was an error parsing this line."
                "Please report this to the package owner.")

    lines[0][0].prefix = first_node_prefix
    return lines



# -*- coding: utf-8 -*-


"""
pep8ify
~~~~~~~~

:copyright: (c) 2012 by Steve Pulec.

"""

__title__ = 'pep8ify'
__version__ = '0.0.5'
__author__ = 'Steve Pulec'
__license__ = 'Apache License 2.0'
__copyright__ = 'Copyright 2012 Steve Pulec'



#!/usr/bin/env python

from os.path import abspath, dirname, join
import sys

from lib2to3.main import main

sys.path.insert(0, abspath(join(dirname(__file__), u"../")))


def _main():
    sys.exit(main("pep8ify.fixes"))

if __name__ == '__main__':
    _main()



from lib2to3.fixer_base import BaseFix
from lib2to3.fixer_util import find_indentation
from lib2to3.pgen2 import token
from lib2to3.pytree import Node, Leaf
from lib2to3.pygram import python_symbols as symbols


class FixImportsOnSeparateLines(BaseFix):
    u'''
    Imports should usually be on separate lines.
    '''

    def match(self, node):
        if (node.type == symbols.simple_stmt and
            node.children[0].type == symbols.import_name and
            node.children[0].children[1].type == symbols.dotted_as_names):
            return node.children[0].children[1].children
        return False

    def transform(self, node, results):
        child_imports = [leaf.value for leaf in results if leaf.type == token.
            NAME]
        current_indentation = find_indentation(node)

        new_nodes = []
        for index, module_name in enumerate(child_imports):
            new_prefix = current_indentation
            if not index:
                # Don't add more indentation if this is the first one
                new_prefix = None
            new_nodes.append(Node(symbols.simple_stmt, [Node(symbols.
                import_name, [Leaf(token.NAME, u'import', prefix=new_prefix),
                Leaf(token.NAME, module_name, prefix=u" ")]), Leaf(token.
                NEWLINE, u'\n')]))

        node.replace(new_nodes)
        node.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols


class FixWhitespaceBeforeParameters(BaseFix):
    u'''
    Avoid extraneous whitespace in the following situations:

    - Immediately before the open parenthesis that starts the argument
      list of a function call.

    - Immediately before the open parenthesis that starts an indexing or
      slicing.
    '''

    def match(self, node):
        if (node.type in (token.LPAR, token.LSQB) and node.parent.type ==
            symbols.trailer):
            return True
        return False

    def transform(self, node, results):
        if node.prefix != u"":
            node.prefix = u""
            node.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.fixer_util import find_indentation
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols

from .utils import (get_whitespace_before_definition, has_parent,
    tuplize_comments)


class FixBlankLines(BaseFix):
    u'''
    Separate top-level function and class definitions with two blank lines.

    Method definitions inside a class are separated by a single blank line.

    Extra blank lines may be used (sparingly) to separate groups of related
    functions.  Blank lines may be omitted between a bunch of related
    one-liners (e.g. a set of dummy implementations).

    Use blank lines in functions, sparingly, to indicate logical sections.
    '''

    def match(self, node):
        # Get classes, non-decorateds funcs, decorators, and simple statements.
        # Ignore decorateds funcs since they will be taken care of with the
        # decorator.
        if (node.type == symbols.funcdef and node.parent.type != symbols.
            decorated or node.type == symbols.classdef or node.type == symbols
            .decorated or node.type == symbols.simple_stmt):
            return True
        return False

    def transform(self, node, results):
        # Sometimes newlines are in prefix of current node, sometimes they're
        # in prefix of the prev sibling
        if node.prefix.count(u'\n'):
            newline_node = node
        else:
            newline_node = get_whitespace_before_definition(node)
            if not newline_node:
                # No previous node, must be the first node.
                return

        if newline_node.type in [token.INDENT, token.NEWLINE]:
            # If the newline_node is an indent or newline, we don't need to
            # worry about fixing indentation since it is not part of the
            # prefix. Dedents do have it as part of the prefix.
            curr_node_indentation = u''
        else:
            curr_node_indentation = find_indentation(node)
        min_lines_between_defs, max_lines_between_defs = (self.
            get_newline_limits(node))
        new_prefix = self.trim_comments(curr_node_indentation, newline_node.
            prefix, min_lines_between_defs, max_lines_between_defs)

        if newline_node.prefix != new_prefix:
            newline_node.prefix = new_prefix
            newline_node.changed()

    def get_newline_limits(self, node):
        if node.type == symbols.simple_stmt or has_parent(node, symbols.
            simple_stmt):
            max_lines_between_defs = 1
            min_lines_between_defs = 0
        elif has_parent(node, symbols.classdef) or has_parent(node, symbols.
            funcdef):
            # If we're inside a definition, only use a single space
            max_lines_between_defs = 1
            min_lines_between_defs = 1
        else:
            # Top-level definition
            max_lines_between_defs = 2
            min_lines_between_defs = 2
        return (min_lines_between_defs, max_lines_between_defs)

    def trim_comments(self, curr_node_indentation, previous_whitespace,
        min_lines_between_defs, max_lines_between_defs):
        before_comments, comments, after_comments = tuplize_comments(
            previous_whitespace)

        if before_comments.count(u"\n") > max_lines_between_defs:
            before_comments = u'\n' * max_lines_between_defs
        if after_comments.count(u"\n") > max_lines_between_defs:
            after_comments = u'\n' * max_lines_between_defs

        if (before_comments.count(u"\n") + after_comments.count(u"\n") >
            max_lines_between_defs):
            if before_comments and after_comments:
                # If there are spaces before and after, trim them down on both
                # sides to either 1 before and 1 after or 0 before and 1 after.
                before_comments = (u'\n' * (min_lines_between_defs - 1) if
                    min_lines_between_defs else u'')
                after_comments = u'\n'

        comment_lines = before_comments.count(u"\n") + after_comments.count(
            u"\n")
        if comment_lines < min_lines_between_defs:
            before_comments += (min_lines_between_defs - comment_lines) * u'\n'
        result = u'%s%s%s' % (before_comments, comments, after_comments)

        # Make sure that the result indenation matches the original indentation
        if result.split(u'\n')[-1] != curr_node_indentation:
            result = u"%s%s" % (result.rstrip(u' '), curr_node_indentation)
        return result



from lib2to3.fixer_base import BaseFix
from lib2to3.pgen2 import token


class FixTrailingWhitespace(BaseFix):
    u'''
    Trailing whitespace is superfluous.
    Except when it occurs as part of a blank line (i.e. the line is
    nothing but whitespace). According to Python docs[1] a line with only
    whitespace is considered a blank line, and is to be ignored. However,
    matching a blank line to its indentation level avoids mistakenly
    terminating a multi-line statement (e.g. class declaration) when
    pasting code into the standard Python interpreter.

    [1] http://docs.python.org/reference/lexical_analysis.html#blank-lines
    '''

    def match(self, node):
        # Newlines can be from a newline token or inside a node prefix
        if node.type == token.NEWLINE or node.prefix.count(u'\n'):
            return True

    def transform(self, node, results):
        if node.prefix.count(u'#'):
            prefix_split = node.prefix.split(u'\n')
            # Rstrip every line except for the last one, since that is the
            # whitespace before this line
            new_prefix = u'\n'.join([line.rstrip(u' \t') for line in
                prefix_split[:-1]] + [prefix_split[-1]])
        else:
            new_prefix = node.prefix.lstrip(u' \t')
        if node.prefix != new_prefix:
            node.prefix = new_prefix
            node.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pgen2 import token

LSTRIP_TOKENS = [token.LPAR, token.LSQB, token.LBRACE]
RSTRIP_TOKENS = [token.RPAR, token.RSQB, token.COLON, token.COMMA, token.SEMI,
    token.RBRACE]
STRIP_TOKENS = RSTRIP_TOKENS + LSTRIP_TOKENS


class FixExtraneousWhitespace(BaseFix):
    u'''
    Avoid extraneous whitespace in the following situations:

    - Immediately inside parentheses, brackets or braces.

    - Immediately before a comma, semicolon, or colon.
    '''

    def match(self, node):
        if node.type in STRIP_TOKENS:
            return True
        return False

    def transform(self, node, results):
        if node.type in LSTRIP_TOKENS and node.get_suffix():
            new_prefix = node.next_sibling.prefix.lstrip(u' \t')
            if node.next_sibling.prefix != new_prefix:
                node.next_sibling.prefix = new_prefix
                node.next_sibling.changed()
        elif node.type in RSTRIP_TOKENS and not node.prefix.count(u'\n'):
            # If the prefix has a newline, this node is the beginning
            # of a newline, no need to do anything.
            new_prefix = node.prefix.rstrip(u' \t')
            if node.prefix != new_prefix:
                node.prefix = new_prefix
                node.changed()



#!/usr/bin/env python

from setuptools import setup
import pep8ify

setup(
    name="pep8ify",
    license='Apache License 2.0',
    version=pep8ify.__version__,
    description="Cleans your python code to conform to pep8",
    author="Steve Pulec",
    author_email="spulec@gmail.com",
    url="https://github.com/spulec/pep8ify",
    packages=["pep8ify", "pep8ify.fixes"],
    entry_points={
        'console_scripts': [
            'pep8ify = pep8ify.pep8ify:_main',
        ],
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ])



from lib2to3.fixer_base import BaseFix
from lib2to3.pgen2 import token


class FixWhitespaceBeforeInlineComment(BaseFix):
    u'''
    Separate inline comments by at least two spaces.

    An inline comment is a comment on the same line as a statement.  Inline
    comments should be separated by at least two spaces from the statement.
    They should start with a # and a single space.
    '''

    _accept_type = token.NEWLINE

    def match(self, node):
        if node.prefix.count(u"#"):
            return True
        return False

    def transform(self, node, results):
        position = node.prefix.find(u"#")
        if position > 2:
            # Already more than two spaces before comment
            whitespace_before, comment_after = node.prefix.split(u"#", 1)
            new_prefix = u"%s# %s" % (whitespace_before, comment_after.lstrip(
                ))
        else:
            new_prefix = u"  # %s" % node.prefix.replace(u"#", u"", 1).lstrip()
        if node.prefix != new_prefix:
            node.prefix = new_prefix
            node.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pytree import Leaf
from lib2to3.pgen2 import token

from .utils import prefix_indent_count


class FixIndentation(BaseFix):
    u'''
    Use 4 spaces per indentation level.

    For really old code that you don't want to mess up, you can continue to
    use 8-space tabs.
    '''

    indents = []
    line_num = 0
    prev_line_indent = 0

    def __init__(self, options, log):
        self.indents = []
        self.line_num = 0
        # This is the indent of the previous line before it was modified
        self.prev_line_indent = 0

        super(FixIndentation, self).__init__(options, log)

    def match(self, node):
        if isinstance(node, Leaf):
            return True
        return False

    def transform(self, node, results):
        if node.type == token.INDENT:
            self.transform_indent(node)
        elif node.type == token.DEDENT:
            self.transform_outdent(node)
        elif self.line_num != node.lineno:
            self.transform_newline(node)

    def transform_indent(self, node):
        self.line_num = node.lineno
        # Indent spacing is stored in the value, node the prefix
        self.prev_line_indent = len(node.value.replace(u'\t', u' ' * 4))
        self.indents.append(self.prev_line_indent)

        new_value = u' ' * 4 * len(self.indents)

        new_prefix = node.prefix
        # Strip any previous newlines since they shouldn't change the comment
        # indent
        comment_indent = node.prefix.strip(u'\n').find(u"#")
        if comment_indent > -1:
            # Determine if we should align the comment with the line before or
            # after
            if comment_indent == node.next_sibling.leaves().next().column:
                # This comment should be aligned with its next_sibling
                new_comment_indent = new_value
            else:
                # This comment should be aligned with the previous indent
                new_comment_indent = u' ' * 4 * (len(self.indents) - 1)

            # Split the lines of comment and prepend them with the new indent
            # value
            new_prefix = ('\n'.join([u"%s%s" % (new_comment_indent, line.
                lstrip()) if line else u'' for line in new_prefix.split('\n')]
                ).rstrip(u' '))

        if node.value != new_value or node.prefix != new_prefix:
            node.value = new_value
            node.prefix = new_prefix
            node.changed()

    def transform_outdent(self, node):
        self.line_num = node.lineno
        self.prev_line_indent = prefix_indent_count(node)

        if node.column:
            # Partial outdent, remove higher indents
            tab_count = node.prefix.count(u'\t')
            if tab_count:
                # If tabs, indent level is number of tabs
                indent_level = tab_count * 4
            else:
                indent_level = node.column
            self.indents = self.indents[:self.indents.index(indent_level) + 1]
            # For OUTDENTS, the indentation is in the last line of the prefix
            self.fix_indent_prefix(node)
        else:
            # Outdent all the way
            self.indents = []

    def transform_newline(self, node):
        self.line_num = node.lineno
        if self.indents:
            self.fix_indent_prefix(node, newline=True)
        else:
            # First line, no need to do anything
            pass

    def fix_indent_prefix(self, node, newline=False):
        if node.prefix:
            new_indent = len(self.indents)
            if newline:
                # Don't reindent continuing lines that are already indented
                # past where they need to be.
                current_indent = prefix_indent_count(node)
                if current_indent > self.prev_line_indent:
                    return

            prefix_lines = node.prefix.split('\n')[:-1]
            prefix_lines.append(u' ' * 4 * new_indent)
            new_prefix = '\n'.join(prefix_lines)
            if node.prefix != new_prefix:
                node.prefix = new_prefix
                node.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.pytree import Leaf


class FixTabs(BaseFix):
    u'''
    For new projects, spaces-only are strongly recommended over tabs.  Most
    editors have features that make this easy to do.
    '''

    def match(self, node):
        if node.prefix.count(u'\t') or (isinstance(node, Leaf)
            and node.value.count(u'\t')):
            return True
        return False

    def transform(self, node, results):
        new_prefix = node.prefix.replace(u'\t', u'    ')
        new_value = node.value.replace(u'\t', u'    ')
        if node.prefix != new_prefix or node.value != new_value:
            node.prefix = new_prefix
            node.value = new_value
            node.changed()



from lib2to3.fixer_base import BaseFix
from lib2to3.fixer_util import find_indentation
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as symbols
from lib2to3.pytree import Node, Leaf


class FixCompoundStatements(BaseFix):
    u'''
    Compound statements (multiple statements on the same line) are
    generally discouraged.

    While sometimes it's okay to put an if/for/while with a small body
    on the same line, never do this for multi-clause statements. Also
    avoid folding such long lines!
    '''

    def match(self, node):
        results = {}
        if (node.prev_sibling and isinstance(node.prev_sibling, Leaf) and node.
            prev_sibling.type == token.COLON and node.type != symbols.suite):
            # If it's inside a lambda definition, subscript, or sliceop, leave
            # it alone
            if node.parent.type in [symbols.lambdef, symbols.subscript,
                symbols.sliceop, symbols.dictsetmaker]:
                pass
            else:
                results["colon"] = True
        if (node.type == symbols.simple_stmt and Leaf(token.SEMI, u';') in node
            .children):
            results["semi"] = True
        return results

    def transform(self, node, results):
        if results.get("colon"):
            node = self.transform_colon(node)
        if results.get("semi"):
            node = self.transform_semi(node)

    def transform_colon(self, node):
        node_copy = node.clone()
        # Strip any whitespace that could have been there
        node_copy.prefix = node_copy.prefix.lstrip()
        old_depth = find_indentation(node)
        new_indent = u'%s%s' % ((u' ' * 4), old_depth)
        new_node = Node(symbols.suite, [Leaf(token.NEWLINE, u'\n'), Leaf(token
            .INDENT, new_indent), node_copy, Leaf(token.DEDENT, u'')])
        node.replace(new_node)
        node.changed()

        # Replace node with new_node in case semi
        return node_copy

    def transform_semi(self, node):
        for child in node.children:
            if child.type == token.SEMI:
                # Strip any whitespace from the next sibling
                if (child.next_sibling.prefix != child.next_sibling.prefix.
                    lstrip()):
                    child.next_sibling.prefix = (child.next_sibling.prefix.
                        lstrip())
                    child.next_sibling.changed()
                # Replace the semi with a newline
                old_depth = find_indentation(child)

                child.replace([Leaf(token.NEWLINE, u'\n'), Leaf(token.INDENT,
                    old_depth)])
                child.changed()
        return node
